#!/usr/bin/env python3

"""
Abstract Base Classes.

pyscp builds most of its functionality on top of three large classes: Wiki,
Page, and Thread. This module contains the abstract base classes for those
three. The ABC-s define the abstact methods that each child must implement,
as well as some common functionality that builds on top of the abstract
methods.

Each class inheriting from the ABC-s must implement its own realization of
the abstract methods, and can also provide additional methods unique to it.

This module also defines the named tuples for simple containers used by the
three core classes, such as Revision or Vote.
"""


###############################################################################
# Module Imports
###############################################################################

import abc
import arrow
import bs4
import collections
import functools
import itertools
import re
import urllib.parse
import logging

import pyscp.utils

###############################################################################
# Global Constants And Variables
###############################################################################

log = logging.getLogger(__name__)

###############################################################################
# Abstract Base Classes
###############################################################################


class Page(metaclass=abc.ABCMeta):
    """
    Page Abstract Base Class.

    Page object are wrappers around individual wiki-pages, and allow simple
    operations with them, such as retrieving the rating or the author.

    Each Page instance is attached to a specific instance of the Wiki class.
    The wiki may be used by the page to retrieve a list of titles or other
    similar wiki-wide information that may be used by the Page to, in turn,
    deduce some information about itself.

    Typically, the Page instances should not be created directly. Instead,
    calling an instance of a Wiki class will creating a Page instance
    attached to that wiki.
    """

    ###########################################################################
    # Special Methods
    ###########################################################################

    def __init__(self, wiki, url):
        self.url = url
        self._wiki = wiki

    def __repr__(self):
        return '{}.{}({}, {})'.format(
            self.__module__, self.__class__.__name__,
            repr(self.url), repr(self._wiki))

    def __eq__(self, other):
        if not hasattr(other, 'url') or not hasattr(other, '_wiki'):
            return False
        return self.url == other.url and self._wiki is other._wiki

    ###########################################################################
    # Abstract Methods
    ###########################################################################

    @property
    @abc.abstractmethod
    def _pdata(self):
        """
        Commonly used data about the page.

        This method should return a tuple, the first three elements of which
        are the id number of the page; the id number of the page's comments
        thread; and the html contents of the page.

        Any additional elements of the tuple are left to the discretion
        of the individual Page implimentations.
        """
        pass

    @property
    @abc.abstractmethod
    def history(self):
        """
        Revision history of the page.

        Should return a sorted list of Revision named tuples.
        """
        pass

    @property
    @abc.abstractmethod
    def votes(self):
        """
        Page votes.

        Should return a list of Vote named tuples.
        """
        pass

    @property
    @abc.abstractmethod
    def tags(self):
        """
        Page tags.

        Should return a set of strings.
        """
        pass

    ###########################################################################
    # Internal Methods
    ###########################################################################

    @property
    def _id(self):
        """Unique ID number of the page."""
        return self._pdata[0]

    @pyscp.utils.cached_property
    def _thread(self):
        """Thread object corresponding to the page's comments thread."""
        return self._wiki.Thread(self._wiki, self._pdata[1])

    @property
    def _raw_title(self):
        """Title as displayed on the page."""
        title = self._soup.find(id='page-title')
        return title.text.strip() if title else ''

    @property
    def _raw_author(self):
        return self.history[0].user

    @property
    def _soup(self):
        """BeautifulSoup of the contents of the page."""
        return bs4.BeautifulSoup(self.html, 'lxml')

    ###########################################################################
    # Properties
    ###########################################################################

    @property
    def html(self):
        """HTML contents of the page."""
        return self._pdata[2]

    @property
    def posts(self):
        """List of the comments made on the page."""
        return self._thread.posts

    @property
    def comments(self):
        """Alias for Page.posts."""
        return self._thread.posts

    @property
    def text(self):
        """Plain text of the page."""
        return self._soup.find(id='page-content').text

    @property
    def wordcount(self):
        """Number of words encountered on the page."""
        return len(re.findall(r"[\w'█_-]+", self.text))

    @property
    def images(self):
        """Number of images dislayed on the page."""
        # TODO: needs more work.
        return [i['src'] for i in self._soup('img')]

    @property
    def name(self):
        return self.url.split('/')[-1]

    @property
    def title(self):
        """
        Title of the page.

        In case of SCP articles, will include the title from the 'series' page.
        """
        try:
            return '{}: {}'.format(
                self._raw_title, self._wiki.titles()[self.url])
        except KeyError:
            return self._raw_title

    @property
    def created(self):
        """When was the page created."""
        return self.history[0].time

    @property
    def metadata(self):
        """
        Return page metadata.

        Authors in this case includes all users related to the creation
        and subsequent maintenance of the page. The values of the dict
        describe the user's relationship to the page.
        """
        data = [i for i in self._wiki.metadata() if i.url == self.url]
        data = {i.user: i for i in data}

        if 'author' not in {i.role for i in data.values()}:
            meta = Metadata(self.url, self._raw_author, 'author', None)
            data[self._raw_author] = meta

        for k, v in data.items():
            if v.role == 'author' and not v.date:
                data[k] = v._replace(date=self.created)

        return data

    @property
    def rating(self):
        """Rating of the page, excluding deleted accounts."""
        return sum(
            v.value for v in self.votes if v.user != '(account deleted)')

    @property
    @pyscp.utils.listify()
    def links(self):
        """
        Other pages linked from this one.

        Returns an ordered list of unique urls. Off-site links or links to
        images are not included.
        """
        unique = set()
        for element in self._soup.select('#page-content a'):
            href = element.get('href', None)
            if (not href or href[0] != '/' or  # bad or absolute link
                    href[-4:] in ('.png', '.jpg', '.gif')):
                continue
            url = self._wiki.site + href.rstrip('|')
            if url not in unique:
                unique.add(url)
                yield url

    @property
    def parent(self):
        """Parent of the current page."""
        if not self.html:
            return None
        breadcrumb = self._soup.select('#breadcrumbs a')
        if breadcrumb:
            return self._wiki.site + breadcrumb[-1]['href']

    @property
    def is_mainlist(self):
        """
        Indicate whether the page is a mainlist scp article.

        This is an scp-wiki exclusive property.
        """
        if 'scp-wiki' not in self._wiki.site:
            return False
        if 'scp' not in self.tags:
            return False
        return bool(re.search(r'/scp-[0-9]{3,4}$', self.url))

    ###########################################################################
    # Methods
    ###########################################################################

    def build_attribution_string(
            self, templates=None, group_templates=None, separator=', ',
            user_formatter=None):
        """
        Create an attribution string based on the page's metadata.

        This is a commonly needed operation. The result should be a nicely
        formatted, human-readable description of who was and is involved with
        the page, and in what role.
        """
        roles = 'author rewrite translator maintainer'.split()

        if not templates:
            templates = {i: '{{user}} ({})'.format(i) for i in roles}

        items = list(self.metadata.values())
        items.sort(key=lambda x: [roles.index(x.role), x.date])

        # group users in the same role on the same date together
        itemdict = collections.OrderedDict()
        for i in items:
            user = user_formatter.format(i.user) if user_formatter else i.user
            key = (i.role, i.date)
            itemdict[key] = itemdict.get(key, []) + [user]

        output = []

        for (role, date), users in itemdict.items():

            hdate = arrow.get(date).humanize() if date else ''

            if group_templates and len(users) > 1:
                output.append(
                    group_templates[role].format(
                        date=date,
                        hdate=hdate,
                        users=', '.join(users[:-1]),
                        last_user=users[-1]))
            else:
                for user in users:
                    output.append(
                        templates[role].format(
                            date=date, hdate=hdate, user=user))

        return separator.join(output)


class Thread(metaclass=abc.ABCMeta):
    """
    Thread Abstract Base Class.

    Thread objects represent individual forum threads. Most pages have a
    corresponding comments thread, accessible via Page._thread.
    """

    def __init__(self, wiki, _id, title=None, description=None):
        self._wiki = wiki
        self._id, self.title, self.description = _id, title, description

    @abc.abstractmethod
    def posts(self):
        """Posts in this thread."""
        pass


class Wiki(metaclass=abc.ABCMeta):
    """
    Wiki Abstract Base Class.

    Wiki objects provide wiki-wide functionality not limited to individual
    pages or threads.
    """

    ###########################################################################
    # Class Attributes
    ###########################################################################

    # should point to the respective Page and Thread classes in each submodule.

    Page = Page
    Thread = Thread

    ###########################################################################
    # Special Methods
    ###########################################################################

    def __init__(self, site):
        parsed = urllib.parse.urlparse(site)
        netloc = parsed.netloc if parsed.netloc else parsed.path
        self.custom_domain = True
        if '.' not in netloc:
            netloc += '.wikidot.com'
            self.custom_domain = False
        elif '.wikidot.com' in netloc:
            self.custom_domain = False
        self.site = urllib.parse.urlunparse(['http', netloc, '', '', '', ''])
        self._title_data = {}

    def __call__(self, name):
        url = name if self.site in name else '{}/{}'.format(self.site, name)
        url = url.replace(' ', '-').replace('_', '-').lower()
        return self.Page(self, url)

    ###########################################################################

    @functools.lru_cache(maxsize=1)
    def metadata(self):
        """
        List page ownership metadata.

        This method is exclusive to the scp-wiki, and is used to fine-tune
        the page ownership information beyond what is possible with Wikidot.
        This allows a single page to have an author different from the user
        who created the zeroth revision of the page, or even have multiple
        users attached to the page in various roles.
        """
        if 'scp-wiki' not in self.site:
            return []
        soup = self('attribution-metadata')._soup
        results = []
        for row in soup('tr')[1:]:
            name, user, type_, date = [i.text.strip() for i in row('td')]
            name = name.lower()
            url = '{}/{}'.format(self.site, name)
            results.append(pyscp.core.Metadata(url, user, type_, date))
        return results

    def _update_titles(self):
        for name in (
                'scp-series', 'scp-series-2', 'scp-series-3', 'scp-series-4', 'scp-series-5', 'scp-series-6',
                'joke-scps', 'scp-ex', 'archived-scps'):
            page = self(name)
            try:
                soup = page._soup
            except:
                continue
            self._title_data[name] = soup

    @functools.lru_cache(maxsize=1)
    @pyscp.utils.ignore(value={})
    @pyscp.utils.log_errors(logger=log.error)
    def titles(self):
        """Dict of url/title pairs for scp articles."""
        if 'scp-wiki' not in self.site:
            return {}

        self._update_titles()

        elems = [i.select('ul > li') for i in self._title_data.values()]
        elems = list(itertools.chain(*elems))
        try:
            elems += list(self('scp-001')._soup(class_='series')[1]('p'))
        except:
            pass

        titles = {}
        for elem in elems:

            sep = ' - ' if ' - ' in elem.text else ', '
            try:
                url1 = self.site + elem.a['href']
                skip, title = elem.text.split(sep, maxsplit=1)
            except (ValueError, TypeError):
                continue

            if title != '[ACCESS DENIED]':
                url2 = self.site + '/' + skip.lower()
                titles[url1] = titles[url2] = title

        return titles

    def list_pages(self, **kwargs):
        """Return pages matching the specified criteria."""
        pages = self._list_pages_parsed(**kwargs)
        author = kwargs.pop('author', None)
        if not author:
            # if 'author' isn't specified, there's no need to check rewrites
            return pages
        include, exclude = set(), set()
        for meta in self.metadata():
            if meta.user == author:
                # if username matches, include regardless of type
                include.add(meta.url)
            elif meta.role == 'author':
                # exclude only if override type is author.
                # if url has author and rewrite author,
                # it will appear in list_pages for both.
                exclude.add(meta.url)
        urls = {p.url for p in pages} | include - exclude
        # if no other options beside author were specified,
        # just return everything we can
        if not kwargs:
            return map(self, sorted(urls))
        # otherwise, retrieve the list of urls without the author parameter
        # to check which urls we should return and in which order
        pages = self._list_pages_parsed(**kwargs)
        return [p for p in pages if p.url in urls]

class User(metaclass=abc.ABCMeta):
    """
    User Abstract Base Class.

    User objects allow you to get a user data.
    """
    
    ###########################################################################
    # Special Methods
    ###########################################################################
    
    def __init__(self, username):
        self.username = username
        self.url = f'http://www.wikidot.com/user:info/{username}'

###############################################################################
# Named Tuple Containers
###############################################################################

nt = collections.namedtuple
Revision = nt('Revision', 'id number user time comment')
Vote = nt('Vote', 'user value')
Post = nt('Post', 'id title content user time parent')
File = nt('File', 'url id name filetype size')
Metadata = nt('Metadata', 'url user role date')
Category = nt('Category', 'id title description size')
Image = nt('Image', 'url source status notes data')
del nt



#!/usr/bin/env python3

"""
Wikidot access classes.

This module contains the classes that facilitate information extraction
and communication with the Wikidot-hosted sites.
"""

###############################################################################
# Module Imports
###############################################################################

from datetime import datetime
import bs4
import functools
import itertools
import logging
import pyscp
import re
import requests

###############################################################################
# Global Constants And Variables
###############################################################################

log = logging.getLogger(__name__)


###############################################################################
# Utility Classes
###############################################################################

class InsistentRequest(requests.Session):
    """Make an auto-retrying request that handles connection loss."""

    def __init__(self, max_attempts=10):
        super().__init__()
        self.max_attempts = max_attempts

    def __repr__(self):
        return '{}(max_attempts={})'.format(
            self.__class__.__name__, self.max_attempts)

    def request(self, method, url, **kwargs):
        logged_kwargs = hide_pass(kwargs)
        logged_kwargs = repr(logged_kwargs) if logged_kwargs else ''
        log.debug('%s: %s %s', method, url, logged_kwargs)

        kwargs.setdefault('timeout', 60)
        kwargs.setdefault('allow_redirects', False)
        for _ in range(self.max_attempts):
            try:
                resp = super().request(method=method, url=url, **kwargs)
            except (
                    requests.ConnectionError,
                    requests.Timeout,
                    requests.exceptions.ChunkedEncodingError):
                continue
            return resp
            if 200 <= resp.status_code < 300:
                return resp
            elif 400 <= resp.status_code < 600:
                continue
        raise requests.ConnectionError(
            'Max retries exceeded with url: {}'.format(url))

    def get(self, url, **kwargs):
        return self.request('GET', url, **kwargs)

    def post(self, url, **kwargs):
        return self.request('POST', url, **kwargs)


def hide_pass(nested_dict):
    result = {}
    for k, v in nested_dict.items():
        if k in ('pass', 'password', 'pasw'):
            result[k] = '********'
        elif isinstance(v, dict):
            result[k] = hide_pass(v)
        else:
            result[k] = v
    return result


###############################################################################


class Page(pyscp.core.Page):
    """Create Page object."""

    def __init__(self, wiki, url):
        super().__init__(wiki, url)
        self._body = {}

    ###########################################################################
    # Internal Methods
    ###########################################################################

    def _module(self, *args, **kwargs):
        """Call Wikidot module."""
        return self._wiki._module(*args, page_id=self._id, **kwargs)

    def _action(self, event, **kwargs):
        """Execute WikiPageAction."""
        return self._module(
            'Empty', action='WikiPageAction', event=event, **kwargs)

    def _vote(self, value):
        """Vote on the page."""
        return self._action(
            'RateAction',
            event='ratePage' if value else 'cancelVote',
            points=value,
            force=True)

    def _flush(self, *names):
        if not hasattr(self, '_cache'):
            return
        self._cache = {k: v for k, v in self._cache.items() if k not in names}

    @pyscp.utils.cached_property
    def _pdata(self):
        data = self._wiki.req.get(self.url).text
        soup = bs4.BeautifulSoup(data, 'lxml')
        return (int(re.search('pageId = ([0-9]+);', data).group(1)),
                parse_element_id(soup.find(id='discuss-button')),
                str(soup.find(id='main-content')),
                {e.text for e in soup.select('.page-tags a')})

    @property
    def _raw_title(self):
        if 'title' in self._body:
            return self._body['title']
        return super()._raw_title

    @property
    def author(self):
        if 'created_by' in self._body:
            return self._body['created_by']
        return super()._raw_author

    ###########################################################################
    # Properties
    ###########################################################################

    @property
    def html(self):
        return self._pdata[2]

    @pyscp.utils.cached_property
    @pyscp.utils.listify()
    def history(self):
        """Return the revision history of the page."""
        data = self._module(
            'history/PageRevisionListModule', page=1, perpage=99999)['body']
        soup = bs4.BeautifulSoup(data, 'lxml')
        for row in reversed(soup('tr')[1:]):
            rev_id = int(row['id'].split('-')[-1])
            cells = row('td')
            number = int(cells[0].text.strip('.'))
            user = cells[4].text
            time = parse_element_time(cells[5])
            comment = cells[6].text if cells[6].text else None
            yield pyscp.core.Revision(rev_id, number, User(user), time, comment)

    @pyscp.utils.cached_property
    def votes(self):
        """Return all votes made on the page."""
        data = self._module('pagerate/WhoRatedPageModule')['body']
        soup = bs4.BeautifulSoup(data, 'lxml')
        spans = [i.text.strip() for i in soup('span')]
        pairs = zip(spans[::2], spans[1::2])
        return [pyscp.core.Vote(u, 1 if v == '+' else -1) for u, v in pairs]

    @property
    def tags(self):
        if 'tags' in self._body:
            return set(self._body['tags'].split())
        return self._pdata[3]

    @property
    def source(self):
        data = self._module('viewsource/ViewSourceModule')['body']
        soup = bs4.BeautifulSoup(data, 'lxml')
        return ''.join([s + '\n' for s in soup.text.replace('\t','').split('\n')[1:]][1:])

    @property
    def created(self):
        if 'created_at' in self._body:
            time = arrow.get(self._body['created_at'], 'DD MMM YYYY HH:mm')
            return time.format('YYYY-MM-DD HH:mm:ss')
        return super().created

    @property
    def rating(self):
        if 'rating' in self._body:
            return int(self._body['rating'])
        return super().rating

    @pyscp.utils.cached_property
    def files(self):
        """List all files attached to the page."""
        data = self._module('files/PageFilesModule')['body']
        soup = bs4.BeautifulSoup(data, 'lxml')
        if not soup.select('table.page-files'):
            return []
        files = soup.select('table.page-files')[0]('tr')[1:]
        parsed = []
        for file in files:
            url = self._wiki.site + file.find('a')['href']
            file_id = int(re.search('event, ([0-9]+)', str(file)).group(1))
            name = file.find('a').text.strip()
            filetype = file('td')[1].text.strip()
            size = file('td')[2].text.strip()
            parsed.append(pyscp.core.File(url, file_id, name, filetype, size))
        return parsed

    ###########################################################################
    # Page-Modifying Methods
    ###########################################################################

    def edit(self, source, title=None, comment=None):
        """Overwrite the page with the new source and title."""
        if title is None:
            title = self._raw_title
        self._flush('html', 'history', 'source')
        wiki_page = self.url.split('/')[-1]
        lock = self._module(
            'edit/PageEditModule',
            mode='page',
            wiki_page=wiki_page,
            force_lock=True)
        return self._action(
            'savePage',
            source=source,
            title=title,
            comments=comment,
            wiki_page=wiki_page,
            lock_id=lock['lock_id'],
            lock_secret=lock['lock_secret'],
            revision_id=lock.get('page_revision_id', None))

    def create(self, source, title, comment=None):
        if not hasattr(self, '_cache'):
            self._cache = {}
        self._cache['_pdata'] = (None, None, None)
        response = self.edit(source, title, comment)
        del self._cache['_pdata']
        return response

    def revert(self, rev_n):
        """Revert the page to a previous revision."""
        self._flush('html', 'history', 'source', 'tags')
        return self._action('revert', revisionId=self.history[rev_n].id)

    def set_tags(self, tags):
        """Replace the tags of the page."""
        res = self._action('saveTags', tags=''.join(tags))
        self._flush('history', '_pdata')
        return res

    def upload(self, name, data):
        url = self._wiki.site + '/default--flow/files__UploadTarget'
        kwargs = dict(
            pageId=self._id,
            page_id=self._id,
            action='FileAction',
            event='uploadFile',
            MAX_FILE_SIZE=52428800)
        response = self._wiki.req.post(
            url,
            data=kwargs,
            files={'userfile': (name, data)},
            headers={'Cookie': self._wiki.cookies})
        response = bs4.BeautifulSoup(response.text, 'lxml')
        status = response.find(id='status').text
        message = response.find(id='message').text
        if status != 'ok' and status != 'file_exists':
            raise RuntimeError(message)
        elif status == 'file_exists':
            self.remove_file(name)
            response = self._wiki.req.post(
                url,
                data=kwargs,
                files={'userfile': (name, data)},
                headers={'Cookie': self._wiki.cookies})
            response = bs4.BeautifulSoup(response.text, 'lxml')
            status = response.find(id='status').text
            message = response.find(id='message').text
            if status != 'ok':
                raise RuntimeError(message)
        return response
    
    def remove_file(self,name):
        files = self.files
        id = -1
        for file in files:
            if file.name == name:
                id = file.id
        self._module('Empty',
            event='deleteFile',
            action='FileAction',
            file_id=id)

    ###########################################################################
    # Voting Methods
    ###########################################################################

    def upvote(self):
        self._vote(1)
        self._flush('votes')

    def downvote(self):
        self._vote(-1)
        self._flush('votes')

    def cancel_vote(self):
        self._vote(0)
        self._flush('votes')


class Thread(pyscp.core.Thread):

    @pyscp.utils.cached_property
    @pyscp.utils.listify()
    def posts(self):
        if self._id is None:
            return
        pages = self._wiki._pager(
            'forum/ForumViewThreadPostsModule', _key='pageNo', t=self._id)
        pages = (bs4.BeautifulSoup(p['body'], 'lxml').body for p in pages)
        pages = (p for p in pages if p)
        posts = (p(class_='post-container', recursive=False) for p in pages)
        posts = itertools.chain.from_iterable(posts)
        for post, parent in crawl_posts(posts):
            post_id = int(post['id'].split('-')[1])
            title = post.find(class_='title').text.strip()
            title = title if title else None
            content = post.find(class_='content')
            content.attrs.clear()
            content = str(content)
            user = post.find(class_='printuser').text
            time = parse_element_time(post)
            yield pyscp.core.Post(post_id, title, content, user, time, parent)

    def new_post(self, source, title=None, parent_id=None):
        return self._wiki._module(
            'Empty',
            threadId=self._id,
            parentId=parent_id,
            title=title,
            source=source,
            action='ForumAction',
            event='savePost')


class Wiki(pyscp.core.Wiki):
    """
    Create a Wiki object.

    This class does not use any of the official Wikidot API, and instead
    relies on sending http post/get requests to internal Wikidot pages and
    parsing the returned data.
    """

    Page = Page
    Thread = Thread
    # Tautology = Tautology

    ###########################################################################
    # Special Methods
    ###########################################################################

    def __init__(self, site):
        super().__init__(site)
        self.req = InsistentRequest()
        self.cookies = 'wikidot_token7=123456;'

    def __repr__(self):
        return '{}.{}({})'.format(
            self.__module__,
            self.__class__.__name__,
            repr(self.site))

    ###########################################################################
    # Internal Methods
    ###########################################################################

    @pyscp.utils.log_errors(log.warning)
    def _module(self, _name, **kwargs):
        """
        Call a Wikidot module.

        This method is responsible for most of the class' functionality.
        Almost all other methods of the class are using _module in one way
        or another.
        """
        response = self.req.post(
            self.site + '/ajax-module-connector.php',
            data=dict(
                pageId=kwargs.get('page_id', None),  # fuck wikidot
                moduleName=_name,
                # token7 can be any 6-digit number, as long as it's the same
                # in the payload and in the cookie
                wikidot_token7='123456',
                **kwargs),
            headers={'Content-Type': 'application/x-www-form-urlencoded;', 'Cookie': self.cookies}).json()
        if response['status'] != 'ok':
            log.error(response)
            raise RuntimeError(response.get('message') or response['status'])
        return response

    def _pager(self, _name, _key, _update=None, **kwargs):
        """Iterate over multi-page module results."""
        first_page = self._module(_name, **kwargs)
        yield first_page
        counter = bs4.BeautifulSoup(
            first_page['body'], 'lxml').find(class_='pager-no')
        if not counter:
            return
        for idx in range(2, int(counter.text.split(' ')[-1]) + 1):
            kwargs.update({_key: idx if _update is None else _update(idx)})
            yield self._module(_name, **kwargs)

    def _list_pages_raw(self, **kwargs):
        """
        Call ListPages module.

        Wikidot's ListPages is an extremely versatile php module that can be
        used to retrieve all sorts of interesting informations, from urls of
        pages created by a given user, and up to full html contents of every
        page on the site.
        """
        yield from self._pager(
            'list/ListPagesModule',
            _key='offset',
            _update=lambda x: 250 * (x - 1),
            perPage=250,
            **kwargs)

    def _list_pages_parsed(self, **kwargs):
        """
        Call ListPages module and parse the results.

        Sets default arguments, parses ListPages body into a namedtuple.
        Returns Page instances with a _body grafted in.
        """
        keys = set(kwargs.pop('body', '').split() + ['fullname'])
        kwargs['module_body'] = '\n'.join(
            map('||{0}||%%{0}%% ||'.format, keys))
        kwargs['created_by'] = kwargs.pop('author', None)
        lists = self._list_pages_raw(**kwargs)
        soups = (bs4.BeautifulSoup(p['body'], 'lxml') for p in lists)
        pages = (s.select('div.list-pages-item') for s in soups)
        pages = itertools.chain.from_iterable(pages)
        for page in pages:
            data = {
                r('td')[0].text: r('td')[1].text.strip() for r in page('tr')}
            page = self(data['fullname'])
            page._body = data
            yield page

    ###########################################################################
    # Public Methods
    ###########################################################################

    def auth(self, username, password):
        """Login to wikidot with the given username/password pair."""
        login = self.req.post(
            'https://www.wikidot.com/default--flow/login__LoginPopupScreen',
            data=dict(
                login=username,
                password=password,
                action='Login2Action',
                event='login'))
        if self.custom_domain == True:
            cd = self.req.get(self.site)
            soup = bs4.BeautifulSoup(cd.text, 'html.parser')
            s = soup.find_all('script')
            for i in s:
                if 'http://www.wikidot.com/default__flow/login__CustomDomainScript' in str(i):
                    siteid = i.get('src')
            redirect = self.req.get(siteid,headers=login.cookies.get_dict())
            redir_url = redirect.text.split('\n')[4].replace("var redir_url = '",'').replace("';",'').replace(' ','') + '&url=' + self.site + '/'
            auth = self.req.get(redir_url)
            for i in auth.cookies.get_dict():
                self.cookies += f'{i}={auth.cookies.get_dict()[i]};'
        else:
            self.cookies += f'WIKIDOT_SESSION_ID={login.cookies["WIKIDOT_SESSION_ID"]};'

    def list_categories(self):
        """Return forum categories."""
        data = self._module('forum/ForumStartModule')['body']
        soup = bs4.BeautifulSoup(data, 'lxml')
        for elem in [e.parent for e in soup(class_='name')]:
            cat_id = parse_element_id(elem.select('.title a')[0])
            title, description, size = [
                elem.find(class_=i).text.strip()
                for i in ('title', 'description', 'threads')]
            yield pyscp.core.Category(
                cat_id, title, description, int(size))

    def list_threads(self, category_id):
        """Return threads in the given category."""
        pages = self._pager(
            'forum/ForumViewCategoryModule', _key='p', c=category_id)
        soups = (bs4.BeautifulSoup(p['body'], 'lxml') for p in pages)
        elems = (s(class_='name') for s in soups)
        for elem in itertools.chain(*elems):
            thread_id = parse_element_id(elem.select('.title a')[0])
            title, description = [
                elem.find(class_=i).text.strip()
                for i in ('title', 'description')]
            yield self.Thread(self, thread_id, title, description)

    def send_pm(self, username, text, title=None):
        lookup = self.req.get(
            'https://www.wikidot.com/quickmodule.php?'
            'module=UserLookupQModule&q=' + username).json()
        if not lookup['users'] or lookup['users'][0]['name'] != username:
            raise ValueError('Username Not Found')
        user_id = lookup['users'][0]['user_id']
        return self.req.post(
            'https://www.wikidot.com/ajax-module-connector.php',
            data=dict(
                moduleName='Empty',
                source=text,
                subject=title,
                to_user_id=user_id,
                action='DashboardMessageAction',
                event='send',
                wikidot_token7='123456'),
            headers={'Content-Type': 'application/x-www-form-urlencoded;', 'Cookie':  self.cookies}).json()
    
    ###########################################################################
    # Properties
    ###########################################################################
    
    @property
    def title(self):
        data = self.req.get(self.site).text
        soup = bs4.BeautifulSoup(data, 'lxml')
        return soup.title.string

    ###########################################################################
    # SCP-Wiki Specific Methods
    ###########################################################################

    @functools.lru_cache(maxsize=1)
    @pyscp.utils.listify()
    def list_images(self):
        if 'scp-wiki' not in self.site:
            return
        base = 'http://scpsandbox2.wikidot.com/image-review-{}'
        urls = [base.format(i) for i in range(1, 36)]
        pages = [self.req.get(u).text for u in urls]
        soups = [bs4.BeautifulSoup(p, 'lxml') for p in pages]
        elems = [s('tr') for s in soups]
        elems = itertools.chain(*elems)
        elems = [e('td') for e in elems]
        elems = [e for e in elems if e]
        for elem in elems:
            url = elem[0].find('img')['src']
            source = elem[2].a['href'] if elem[2]('a') else None
            status, notes = [elem[i].text for i in (3, 4)]
            status, notes = [i if i else None for i in (status, notes)]
            yield pyscp.core.Image(url, source, status, notes, None)

class User(pyscp.core.User):
    """
    Create a User object.
    """

    ###########################################################################
    # Special Methods
    ###########################################################################
    
    def __init__(self, username):
        super().__init__(username)
        self.req = InsistentRequest()
    
    def __repr__(self):
        return '{}.{}({})'.format(
            self.__module__,
            self.__class__.__name__,
            repr(self.username))
    
    ###########################################################################
    # Internal Methods
    ###########################################################################
    
    @pyscp.utils.log_errors(log.warning)
    def _module(self, _name, **kwargs):
        response = self.req.post(
            'http://www.wikidot.com/ajax-module-connector.php',
            data=dict(user_id=self.id,
                moduleName=_name,
                wikidot_token7='123456',
                **kwargs),
            headers={'Content-Type': 'application/x-www-form-urlencoded;', 'Cookie': 'wikidot_token7=123456;'}).json()
        if response['status'] != 'ok':
            log.error(response)
            raise RuntimeError(response.get('message') or response['status'])
        return response
    
    @pyscp.utils.cached_property
    def _pdata(self):
        data = self.req.get(self.url).text
        soup = bs4.BeautifulSoup(data, 'lxml')
        return self.parse_data(soup)
    
    @pyscp.utils.log_errors(log.warning)
    def parse_data(self, soup):
        content = soup.find(id='page-content')
        time = parse_element_time(content)
        level = 'none'
        if 'none' in content:
            level = 'none'
        elif 'low' in content:
            level = 'low'
        elif 'medium' in content:
            level = 'medium'
        elif 'high' in content and 'very' not in content:
            level = 'high'
        elif 'very high' in content:
            level = 'very high'
        elif 'guru' in content:
            level = 'guru'
        utype = 'free'
        if 'free' in content:
            utype = 'free'
        elif 'Pro' in content:
            utype = 'Pro'
        return {'name': self.username, 'time': time, 'type': utype, 'level': level}
    
    @property
    def id(self):
        data = self.req.get(self.url).text
        soup = bs4.BeautifulSoup(data, 'lxml')
        return int(re.search('userId = ([0-9]+);', data).group(1))
    
    ###########################################################################
    # Properties
    ###########################################################################
    
    @property
    def join_time(self):
        return self._pdata['time']
    
    @property
    def type(self):
        return self._pdata['type']
    
    @property
    def karma(self):
        return self._pdata['level']
    
    @property
    def member(self):
        data = self._module('userinfo/UserInfoMemberOfModule')['body']
        soup = bs4.BeautifulSoup(data, 'lxml')
        result = []
        for e in soup.find_all('a'):
            result.append(Wiki(e.get('href')))
        return result
    
    @property
    def moderator(self):
        data = self._module('userinfo/UserInfoModeratorOfModule')['body']
        soup = bs4.BeautifulSoup(data, 'lxml')
        result = []
        for e in soup.find_all('a'):
            result.append(Wiki(e.get('href')))
        return result
    
    @property
    def admin(self):
        data = self._module('userinfo/UserInfoAdminOfModule')['body']
        soup = bs4.BeautifulSoup(data, 'lxml')
        result = []
        for e in soup.find_all('a'):
            result.append(Wiki(e.get('href')))
        return result

###############################################################################


@pyscp.utils.ignore((IndexError, TypeError))
def parse_element_id(element):
    """Extract the id number from the link."""
    return int(element['href'].split('/')[2].split('-')[1])


def parse_element_time(element):
    """Extract and format time from an html element."""
    unixtime = int(element.find(class_='odate')['class'][1].split('_')[1])
    return datetime.utcfromtimestamp(unixtime).strftime('%Y-%m-%d %H:%M:%S')


def crawl_posts(post_containers, parent=None):
    """
    Retrieve posts from the comment tree.

    For each post-container in the given list, returns a tuple of
    (post, parent). Then recurses onto all the post-container children
    of the current post-container.
    """
    for container in post_containers:
        yield container.find(class_='post'), parent
        yield from crawl_posts(
            container(class_='post-container', recursive=False),
            int(container['id'].split('-')[1]))



#!/usr/bin/env python3

###############################################################################
# Module Imports
###############################################################################

import concurrent.futures
import logging
import peewee
import queue

from itertools import islice

###############################################################################
# Global Constants And Variables
###############################################################################

log = logging.getLogger('pyscp.orm')
pool = concurrent.futures.ThreadPoolExecutor(max_workers=1)
queue = queue.Queue()


def queue_execution(fn, args=(), kw={}):
    queue.put(dict(fn=fn, args=args, kw=kw))
    pool.submit(async_write)

###############################################################################
# Database ORM Classes
###############################################################################

db = peewee.Proxy()


class BaseModel(peewee.Model):

    class Meta:
        database = db

    @classmethod
    def create(cls, **kw):
        queue_execution(fn=super().create, kw=kw)

    @classmethod
    def create_table(cls):
        if not hasattr(cls, '_id_cache'):
            cls._id_cache = []
        queue_execution(fn=super().create_table, args=(True,))

    @classmethod
    def insert_many(cls, data):
        data_iter = iter(data)
        chunk = list(islice(data_iter, 500))
        while chunk:
            queue_execution(
                fn=lambda x: super(BaseModel, cls).insert_many(x).execute(),
                args=(chunk, ))
            chunk = list(islice(data_iter, 500))

    @classmethod
    def convert_to_id(cls, data, key='user'):
        for row in data:
            if row[key] not in cls._id_cache:
                cls._id_cache.append(row[key])
            row[key] = cls._id_cache.index(row[key]) + 1
            yield row

    @classmethod
    def write_ids(cls, field_name):
        cls.insert_many([
            {'id': cls._id_cache.index(value) + 1, field_name: value}
            for value in set(cls._id_cache)])
        cls._id_cache.clear()


class ForumCategory(BaseModel):
    title = peewee.CharField()
    description = peewee.TextField()


class ForumThread(BaseModel):
    category = peewee.ForeignKeyField(ForumCategory, null=True)
    title = peewee.CharField(null=True)
    description = peewee.TextField(null=True)


class Page(BaseModel):
    url = peewee.CharField(unique=True)
    html = peewee.TextField()
    thread = peewee.ForeignKeyField(
        ForumThread, related_name='page', null=True)


class User(BaseModel):
    name = peewee.CharField(unique=True)


class Revision(BaseModel):
    page = peewee.ForeignKeyField(Page, related_name='revisions', index=True)
    user = peewee.ForeignKeyField(User, related_name='revisions', index=True)
    number = peewee.IntegerField()
    time = peewee.DateTimeField()
    comment = peewee.CharField(null=True)


class Vote(BaseModel):
    page = peewee.ForeignKeyField(Page, related_name='votes', index=True)
    user = peewee.ForeignKeyField(User, related_name='votes', index=True)
    value = peewee.IntegerField()


class ForumPost(BaseModel):
    thread = peewee.ForeignKeyField(
        ForumThread, related_name='posts', index=True)
    user = peewee.ForeignKeyField(User, related_name='posts', index=True)
    parent = peewee.ForeignKeyField('self', null=True)
    title = peewee.CharField(null=True)
    time = peewee.DateTimeField()
    content = peewee.TextField()


class Tag(BaseModel):
    name = peewee.CharField(unique=True)


class PageTag(BaseModel):
    page = peewee.ForeignKeyField(Page, related_name='tags', index=True)
    tag = peewee.ForeignKeyField(Tag, related_name='pages', index=True)


class OverrideType(BaseModel):
    name = peewee.CharField(unique=True)


class Override(BaseModel):
    url = peewee.ForeignKeyField(Page, to_field=Page.url, index=True)
    user = peewee.ForeignKeyField(User, index=True)
    type = peewee.ForeignKeyField(OverrideType)


class ImageStatus(BaseModel):
    name = peewee.CharField(unique=True)


class Image(BaseModel):
    url = peewee.CharField(unique=True)
    source = peewee.CharField()
    data = peewee.BlobField()
    status = peewee.ForeignKeyField(ImageStatus)
    notes = peewee.TextField(null=True)

###############################################################################
# Helper Functions
###############################################################################


def async_write(buffer=[]):
    item = queue.get()
    buffer.append(item)
    if len(buffer) > 500 or queue.empty():
        log.debug('Processing {} queue items.'.format(len(buffer)))
        with db.transaction():
            write_buffer(buffer)
        buffer.clear()


def write_buffer(buffer):
    for item in buffer:
        try:
            item['fn'](*item.get('args', ()), **item.get('kw', {}))
        except:
            log.exception(
                'Exception while processing queue item: {}'
                .format(item))
        queue.task_done()


def create_tables(*tables):
    for table in tables:
        eval(table).create_table()


def connect(dbpath):
    log.info('Connecting to the database at {}'.format(dbpath))
    db.initialize(peewee.SqliteDatabase(dbpath))
    db.connect()


###############################################################################
# Macros
###############################################################################


def votes_by_user(user):
    up, down = [], []
    for vote in (Vote.select().join(User).where(User.name == user)):
        if vote.value == 1:
            up.append(vote.page.url)
        else:
            down.append(vote.page.url)
    return {'+': up, '-': down}



#!/usr/bin/env python3

###############################################################################
# Module Imports
###############################################################################

import logging
import re
import time
import threading
import signal
import functools
import inspect

###############################################################################
# Decorators
###############################################################################

###############################################################################
# Decorator decorator is a simplified version of the code from the funcy lib.
# https://github.com/Suor/funcy
###############################################################################


class Call:

    def __init__(self, func, args, kwargs):
        self.func, self.args, self.kwargs = func, args, kwargs

    def __call__(self):
        return self.func(*self.args, **self.kwargs)


def decorator(deco):
    spec = inspect.getargspec(deco)
    if len(spec.args) > 1 or spec.varargs or spec.keywords:
        @functools.wraps(deco)
        def _fab(*dargs, **dkwargs):
            return make_decorator(deco, *dargs, **dkwargs)
        return _fab
    else:
        return functools.wraps(deco)(make_decorator(deco))


def make_decorator(deco, *dargs, **dkwargs):
    def _decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            call = Call(func, args, kwargs)
            return deco(call, *dargs, **dkwargs)
        return wrapper
    return _decorator

###############################################################################


@decorator
def listify(call, wrapper=list):
    return wrapper(call())


@decorator
def morph(call, catch_exc, raise_exc):
    try:
        return call()
    except catch_exc as error:
        raise raise_exc(error) from error


@decorator
def ignore(call, error=Exception, value=None):
    try:
        return call()
    except error:
        return value


@decorator
def log_errors(call, logger=print):
    try:
        return call()
    except Exception as error:
        logger(error)
        raise(error)


@decorator
def decochain(call, *decs):
    fn = call.func
    for dec in reversed(decs):
        fn = dec(fn)
    return fn(*call.args, **call.kwargs)


class cached_property:

    def __init__(self, func):
        self.func = func
        functools.update_wrapper(self, func)

    def __get__(self, obj, cls):
        if not hasattr(obj, '_cache'):
            obj._cache = {}
        if self.func.__name__ not in obj._cache:
            obj._cache[self.func.__name__] = self.func(obj)
        return obj._cache[self.func.__name__]

###############################################################################


def split(text, delimeters):
    pattern = '|'.join(map(re.escape, delimeters))
    return re.split(pattern, text)


class ProgressBar:

    def __init__(self, title, max_value):
        self.title = title
        self.max_value = max_value
        self.value = 0
        signal.signal(signal.SIGINT, self.exit)

    def start(self):
        self.finished = False
        self.time_started = time.time()
        threading.Thread(target=self.run).start()

    def update(self):
        print(self.line() + '\r', end='')

    def line(self):
        filled = 40 * self.value / self.max_value
        parts = ' ▏▎▍▌▋▊▉'
        current = int(filled * len(parts)) % len(parts)
        bar = '█' * int(filled) + parts[current] + ' ' * 40
        tm = time.gmtime(time.time() - self.time_started)
        return '{} |{}| {:>3}% ({}:{:02}:{:02})   '.format(
            self.title,
            bar[:40],
            100 * self.value // self.max_value,
            tm.tm_hour, tm.tm_min, tm.tm_sec)

    def run(self):
        while not self.finished:
            self.update()
            time.sleep(1)

    def stop(self):
        self.finished = True
        print(self.line())

    def exit(self, signum, frame):
        self.stop()
        raise KeyboardInterrupt


def pbar(it, title=None, max=None):
    max = len(it) if max is None else max
    title = '' if title is None else title + ' '
    bar = ProgressBar(title, max)
    bar.start()
    for i in it:
        yield i
        bar.value += 1
        bar.update()
    bar.stop()

###############################################################################


class LogCount:

    def __init__(self):
        self.count = 1

    def filter(self, record):
        record.count = self.count
        self.count += 1
        return True


def log_sql_debug():
    logger = logging.getLogger('peewee')
    logger.setLevel(logging.DEBUG)
    logger.addFilter(LogCount())
    term = logging.StreamHandler()
    term.setFormatter(logging.Formatter('{count} {message}', style='{'))
    logger.addHandler(term)


def default_logging(debug=False):
    term = logging.StreamHandler()
    file = logging.FileHandler('pyscp.log', mode='a', delay=True)
    if debug:
        term.setLevel(logging.DEBUG)
        file.setLevel(logging.DEBUG)
    else:
        term.setLevel(logging.INFO)
        file.setLevel(logging.INFO)
    term.setFormatter(logging.Formatter('{message}', style='{'))
    file.setFormatter(
        logging.Formatter('{asctime} {levelname:8s} {message}', style='{'))
    logger = logging.getLogger('pyscp')
    logger.setLevel(logging.DEBUG)
    logger.addHandler(term)
    logger.addHandler(file)

###############################################################################



#!/usr/bin/env python3

"""
Snapshot access classes.

This module contains the classes that facilitate information extraction
and communication with the sqlite Snapshots.
"""

###############################################################################
# Module Imports
###############################################################################

import bs4
import concurrent.futures
import functools
import itertools
import logging
import operator
import pathlib
import re
import requests

from pyscp import core, orm, utils

###############################################################################
# Global Constants And Variables
###############################################################################

log = logging.getLogger(__name__)

###############################################################################


class Page(core.Page):
    """Page object."""

    ###########################################################################
    # Internal Methods
    ###########################################################################

    def _query(self, ptable, stable='User'):
        """Generate SQL queries used to retrieve data."""
        pt, st = [getattr(orm, i) for i in (ptable, stable)]
        return pt.select(pt, st.name).join(st).where(pt.page == self._id)

    @utils.cached_property
    def _pdata(self):
        """Preload the ids and contents of the page."""
        pdata = orm.Page.get(orm.Page.url == self.url)
        return pdata.id, pdata._data['thread'], pdata.html

    ###########################################################################
    # Properties
    ###########################################################################

    @property
    def html(self):
        """Return HTML contents of the page."""
        return self._pdata[2]

    @utils.cached_property
    def history(self):
        """Return the revisions of the page."""
        revs = self._query('Revision')
        revs = sorted(revs, key=lambda x: x.number)
        return [core.Revision(
                r.id, r.number, r.user.name, str(r.time), r.comment)
                for r in revs]

    @utils.cached_property
    def votes(self):
        """Return all votes made on the page."""
        return [core.Vote(v.user.name, v.value)
                for v in self._query('Vote')]

    @utils.cached_property
    def tags(self):
        """Return the set of tags with which the page is tagged."""
        return {pt.tag.name for pt in self._query('PageTag', 'Tag')}


class Thread(core.Thread):
    """Discussion/forum thread."""

    @utils.cached_property
    def posts(self):
        """Post objects belonging to this thread."""
        fp = orm.ForumPost
        us = orm.User
        query = fp.select(fp, us.name).join(us).where(fp.thread == self._id)
        return [core.Post(
                p.id, p.title, p.content, p.user.name,
                str(p.time), p._data['parent'])
                for p in query]


class Wiki(core.Wiki):
    """Snapshot of a Wikidot website."""

    Page = Page
    Thread = Thread
    # Tautology = Tautology

    ###########################################################################
    # Special Methods
    ###########################################################################

    def __init__(self, site, dbpath):
        """Create wiki instance."""
        super().__init__(site)
        if not pathlib.Path(dbpath).exists():
            raise FileNotFoundError(dbpath)
        self.dbpath = dbpath
        orm.connect(dbpath)

    def __repr__(self):
        """Pretty-print current instance."""
        return '{}.{}({}, {})'.format(
            self.__module__,
            self.__class__.__qualname__,
            repr(self.site),
            repr(self.dbpath))

    ###########################################################################
    # Internal Methods
    ###########################################################################

    @staticmethod
    def _filter_author(author):
        return (orm.Page.select(orm.Page.url)
                .join(orm.Revision).join(orm.User)
                .where(orm.Revision.number == 0)
                .where(orm.User.name == author))

    @staticmethod
    def _filter_tag(tag):
        return (orm.Page.select(orm.Page.url)
                .join(orm.PageTag).join(orm.Tag)
                .where(orm.Tag.name == tag))

    @staticmethod
    def _get_operator(string):
        symbol, *values = re.split(r'(\d+)', string)
        opdict = {
            '>': 'gt', '<': 'lt', '>=': 'ge', '<=': 'le', '=': 'eq', '': 'eq'}
        if symbol not in opdict:
            raise ValueError
        return getattr(operator, opdict[symbol]), values

    def _filter_rating(self, rating):
        compare, values = self._get_operator(rating)
        rating = int(values[0])
        return (orm.Page.select(orm.Page.url)
                .join(orm.Vote).group_by(orm.Page.url)
                .having(compare(orm.peewee.fn.sum(orm.Vote.value), rating)))

    def _filter_created(self, created):
        compare, values = self._get_operator(created)
        date = '-'.join(values[::2])
        return (orm.Page.select(orm.Page.url)
                .join(orm.Revision).where(orm.Revision.number == 0)
                .group_by(orm.Page.url)
                .having(compare(
                    orm.peewee.fn.substr(orm.Revision.time, 1, len(date)),
                    date)))

    def _list_pages_parsed(self, **kwargs):
        query = orm.Page.select(orm.Page.url)
        keys = ('author', 'tag', 'rating', 'created')
        keys = [k for k in keys if k in kwargs]
        for k in keys:
            query = query & getattr(self, '_filter_' + k)(kwargs[k])
        if 'limit' in kwargs:
            query = query.limit(kwargs['limit'])
        return map(self, [p.url for p in query])

    ###########################################################################
    # SCP-Wiki Specific Methods
    ###########################################################################

    @functools.lru_cache(maxsize=1)
    def list_images(self):
        """Image metadata."""
        query = (
            orm.Image.select(orm.Image, orm.ImageStatus.name)
            .join(orm.ImageStatus))
        return [core.Image(r.url, r.source, r.status.name, r.notes, r.data)
                for r in query]

###############################################################################


class SnapshotCreator:
    """
    Create a snapshot of a wikidot site.

    This class uses WikidotConnector to iterate over all the pages of a site,
    and save the html content, revision history, votes, and the discussion
    of each to a sqlite database. Optionally, standalone forum threads can be
    saved too.

    In case of the scp-wiki, some additional information is saved:
    images for which their CC status has been confirmed, and info about
    overwriting page authorship.

    In general, this class will not save images hosted on the site that is
    being saved. Only the html content, discussions, and revision/vote
    metadata is saved.
    """

    def __init__(self, dbpath):
        """Create an instance."""
        if pathlib.Path(dbpath).exists():
            raise FileExistsError(dbpath)
        orm.connect(dbpath)
        self.pool = concurrent.futures.ThreadPoolExecutor(max_workers=20)

    def take_snapshot(self, wiki, forums=False):
        """Take new snapshot."""
        self.wiki = wiki
        self._save_all_pages()
        if forums:
            self._save_forums()
        if 'scp-wiki' in self.wiki.site:
            self._save_meta()
        orm.queue.join()
        self._save_cache()
        orm.queue.join()
        log.info('Snapshot succesfully taken.')

    def _save_all_pages(self):
        """Iterate over the site pages, call _save_page for each."""
        orm.create_tables(
            'Page', 'Revision', 'Vote', 'ForumPost',
            'PageTag', 'ForumThread', 'User', 'Tag')
        count = next(
            self.wiki.list_pages(body='total', limit=1))._body['total']
        bar = utils.ProgressBar('SAVING PAGES'.ljust(20), int(count))
        bar.start()
        for _ in self.pool.map(self._save_page, self.wiki.list_pages()):
            bar.value += 1
        bar.stop()

    @utils.ignore(requests.HTTPError)
    def _save_page(self, page):
        """Download contents, revisions, votes and discussion of the page."""
        orm.Page.create(
            id=page._id, url=page.url, thread=page._thread._id, html=page.html)

        revisions = orm.User.convert_to_id(i._asdict() for i in page.history)
        votes = orm.User.convert_to_id(i._asdict() for i in page.votes)
        tags = [{'tag': t} for t in page.tags]
        tags = orm.Tag.convert_to_id(tags, key='tag')

        def _insert(table, data):
            table.insert_many(dict(i, page=page._id) for i in data)

        _insert(orm.Revision, revisions)
        _insert(orm.Vote, votes)
        _insert(orm.PageTag, tags)

        self._save_thread(page._thread)

    def _save_forums(self):
        """Download and save standalone forum threads."""
        orm.create_tables(
            'ForumPost', 'ForumThread', 'ForumCategory', 'User')
        cats = self.wiki.list_categories()
        cats = [i for i in cats if i.title != 'Per page discussions']
        orm.ForumCategory.insert_many(dict(
            id=c.id,
            title=c.title,
            description=c.description) for c in cats)
        total_size = sum(c.size for c in cats)
        bar = utils.ProgressBar('SAVING FORUM THREADS', total_size)
        bar.start()
        for cat in cats:
            threads = set(self.wiki.list_threads(cat.id))
            c_id = itertools.repeat(cat.id)
            for _ in self.pool.map(self._save_thread, threads, c_id):
                bar.value += 1
        bar.stop()

    def _save_thread(self, thread, c_id=None):
        orm.ForumThread.create(
            category=c_id, id=thread._id,
            title=thread.title, description=thread.description)
        posts = orm.User.convert_to_id([i._asdict() for i in thread.posts])
        orm.ForumPost.insert_many(
            dict(p, thread=thread._id) for p in posts)

    def _save_meta(self):
        orm.create_tables(
            'Image', 'ImageStatus')
        licenses = {
            'PERMISSION GRANTED', 'BY-NC-SA CC', 'BY-SA CC', 'PUBLIC DOMAIN'}
        images = [i for i in self.wiki.list_images() if i.status in licenses]
        self.ibar = utils.ProgressBar(
            'SAVING IMAGES'.ljust(20), len(images))
        self.ibar.start()
        data = list(self.pool.map(self._save_image, images))
        self.ibar.stop()
        images = orm.ImageStatus.convert_to_id(
            [i._asdict() for i in images], key='status')
        orm.Image.insert_many(
            dict(i, data=d) for i, d in zip(images, data) if d)

    @utils.ignore(requests.RequestException)
    def _save_image(self, image):
        self.ibar.value += 1
        if not image.source:
            log.info('Image source not specified: ' + image.url)
            return
        return self.wiki.req.get(image.url, allow_redirects=True).content

    def _save_cache(self):
        for table in orm.User, orm.Tag, orm.OverrideType, orm.ImageStatus:
            if hasattr(table, '_id_cache') and table._id_cache:
                table.write_ids('name')



import setuptools

with open('requirements.txt') as f:
  requirements = f.read().splitlines()

with open('README.md', encoding="utf8") as f:
    readme = f.read()

setuptools.setup(
    name='pyscp',
    version='1.1.12',
    description='Python API and utilities for the www.scpwiki.com website.',
    long_description=readme,
    long_description_content_type="text/markdown",
    url='https://github.com/MrNereof/pyscp/',
    author='MrNereof',
    author_email='mrnereof@gmail.com',
    python_requires='>=3.4',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'],
    keywords=['scp', 'wikidot'],
    packages=['pyscp'],
    install_requires=requirements,
)
